/**
 * Illustrates global memory accesses
 *
 * Lecture notes:
 * 1) Understand the global memory access pattern of transpose_naive. Can we improve it? 
 * 2) Profile the naive version. Have a look at "Global store efficiency"
 * 3) Profile sm version. Have a look at "Shared Memory Replay Overhead", which is an indicator for bank conflicts.
 * 4) Remove bank conflict
 */

#include <cuda_runtime.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "../misc/cuda_util.h"

#define BLOCKSIZE 128
#define NUM_BLOCKS 128

/// f(x) = 3.3 x**3 + 2.2 * x**2 + 1.1 * x + 0.1
__device__ float f(float x)
{
   return 3.3f * x * x * x + 2.2f * x * x + 1.1f * x + 0.1f;
}

__global__ void func(float * x, int n)
{
   int idx = blockIdx.x * blockDim.x + threadIdx.x;

   int i;
   for(; idx < n ; idx +=  gridDim.x * blockDim.x){
      float tmp = x[idx];
      for(i=0;i < 20; ++i){
         tmp += f(tmp);
      }
      x[idx] = tmp;
   }
}


int main(int argc, char** argv)
{
   int i;
   int n= 1024*1024*10;
   if(argc > 1)
      n = atoi(argv[1]);

   int size = sizeof(float) *n;
   printf("Allocates %f MB\n", ((float)size)/1024/1024);


   //allocate host memory
   float *x_h,*y_h;
   HANDLE_ERROR( cudaMallocHost(&x_h, size) ); //allocate page-locked memory
   HANDLE_ERROR( cudaMallocHost(&y_h, size) ); //allocate page-locked memory


   //init x and y with random numbers in [0,1]
   for(i=0;i < n; ++i){
      x_h[i] = ((float)rand())/RAND_MAX;
   }

   /**********************************************
    * Get Device properties 
    ***********************************************/
   cudaDeviceProp deviceProp;
   cudaGetDeviceProperties(&deviceProp, 0);

   if(!deviceProp.deviceOverlap)
      printf("WARNING: Your GPU does not support concurrent copy and kernel execution!\n");
   else if(deviceProp.asyncEngineCount == 1)
      printf("WARNING: Your GPU does only have 1 copy engine!\n");


   //allocate device memory
   float *x_d;
   HANDLE_ERROR( cudaMalloc(&x_d, size) );


   /**********************************************
    * Sync memcopy
    ***********************************************/
   float time;
   cudaEvent_t start, stop; //used for timings
   HANDLE_ERROR( cudaEventCreate(&start) );
   HANDLE_ERROR( cudaEventCreate(&stop) );
   HANDLE_ERROR( cudaEventRecord(start, 0) ); 

   //copy data from host to device
   HANDLE_ERROR( cudaMemcpy(x_d, x_h, size, cudaMemcpyHostToDevice) ); 

   func<<<NUM_BLOCKS,BLOCKSIZE>>>(x_d, n);

   //copy data from device to Host
   HANDLE_ERROR( cudaMemcpy(y_h, x_d, size, cudaMemcpyDeviceToHost) ); 

   HANDLE_ERROR( cudaEventRecord(stop, 0) );
   HANDLE_ERROR( cudaEventSynchronize(stop) );
   HANDLE_ERROR( cudaEventElapsedTime(&time, start, stop) );
   printf("Time (no streams):  %5.3f ms \n", time);


   /**********************************************
    * ASync memcopy in stages
    ***********************************************/
   HANDLE_ERROR( cudaEventCreate(&start) );
   HANDLE_ERROR( cudaEventCreate(&stop) );
   HANDLE_ERROR( cudaEventRecord(start, 0) ); 

   int nStreams = 8;
   int n_per_stream = n / nStreams;
   int size_per_stream = size / nStreams;

   //create cuda sterams
   cudaStream_t* streams = (cudaStream_t*) malloc(sizeof(cudaStream_t) * nStreams);
   for(i=0;i < nStreams; ++i){
      HANDLE_ERROR( cudaStreamCreate(&streams[i]) );
   }

   if(!(deviceProp.asyncEngineCount > 1)){
      //if you would run this piece of code with only 1 copyEngine, 
      //your code would run sequentially for devices with CC <= 3.0

      for(i=0;i < nStreams; ++i){

         // HTD
         HANDLE_ERROR( cudaMemcpyAsync(&x_d[i * n_per_stream], &x_h[i * n_per_stream],
                  size_per_stream, cudaMemcpyHostToDevice, streams[i]) ); 

         // KERNEL
         func<<<NUM_BLOCKS,BLOCKSIZE,0 , streams[i]>>>(&x_d[i * n_per_stream], n_per_stream);

         // DTH
         HANDLE_ERROR( cudaMemcpyAsync(&y_h[i * n_per_stream], &x_d[i * n_per_stream], size_per_stream, cudaMemcpyDeviceToHost, streams[i]) ); 

      }
   }else{
      //execution order matters! This piece of code avoids serialization

      for(i=0;i < nStreams; ++i){

         // HTD
         HANDLE_ERROR( cudaMemcpyAsync(&x_d[i * n_per_stream], &x_h[i * n_per_stream],
                  size_per_stream, cudaMemcpyHostToDevice, streams[i]) ); 

         // KERNEL
         func<<<NUM_BLOCKS,BLOCKSIZE,0 , streams[i]>>>(&x_d[i * n_per_stream], n_per_stream);

      }

      // DTH
      for(i=0;i < nStreams; ++i){
         HANDLE_ERROR( cudaMemcpyAsync(&y_h[i * n_per_stream], &x_d[i * n_per_stream], size_per_stream, cudaMemcpyDeviceToHost, streams[i]) ); 
      }
   }

   cudaDeviceSynchronize(); //Required to wait for all streams to complete

   HANDLE_ERROR( cudaEventRecord(stop, 0) ); //insert stop event into stream 0
   HANDLE_ERROR( cudaEventSynchronize(stop) ); //wait till the stop event has occured
   HANDLE_ERROR( cudaEventElapsedTime(&time, start, stop) ); //take time difference between start and stop event
   printf("Time (streams):  %5.3f ms \n", time);

   for(i=0;i < nStreams; ++i)
      cudaStreamDestroy(streams[i]);


   //free memory
   cudaFree(x_h); 
   cudaFree(x_d);

   return 0;
}




